import boto3
import json

def get_object(bucket, key):
    s3 = boto3.client('s3')
    return s3.get_object(Bucket=bucket, Key=key)

def get_content(s3_object):
    return s3_object.get('Body').read().decode('utf-8')

def get_json_content(s3_object):
    return json.loads(get_content(s3_object))


import boto3
import json
import requests

class Search(object):
    def __init__(self, domain, endpoint):
        self._domain = domain 
        self._endpoint = endpoint

    def search(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError

    def put_documents(self):
        raise NotImplementedError

    def delete_document(self):
        raise NotImplementedError

    def get_index(self):
        raise NotImplementedError

    def put_index(self):
        raise NotImplementedError

    def delete_index(self):
        raise NotImplementedError

    def get_type(self):
        raise NotImplementedError


class ElasticSearch(Search):
    """
    To further improve this, add cloudwatch here for non-200 status code response
    for delete/put methods
    """
    def __init__(self, domain, endpoint):
        super().__init__(domain, endpoint)

    def search(self, keywords, index, type='', keyward_operator='OR', fields=[], filters={}, **kwargs):
        url = ''
        q = ''

        # if type parameter provided the search scope is set to type level
        # otherwise index level
        if type:
            url = '{}/{}/{}/_search'.format(self._endpoint, index, type)
        else:
            url = '{}/{}/_search'.format(self._endpoint, index)


        if kwargs.get('fuzzy', False):
            q = ' {} '.format(keyward_operator).join(['*{}*'.format(item) for item in keywords])
        else:
            q = ' {} '.format(keyward_operator).join(keywords)

        search_criterias={'query': {'query_string':{'default_operator': keyward_operator,
                                                    'query': q}}} 
       
        #fields parameter will not work on default fields by ES, because
        #the analyzer will not recognize it
        if fields:
            search_criterias['query']['query_string'].update({'fields': fields})

        if filters:
            filter_str = ''
            for key, val in filters.items():
                """
                Here's the tricky part
                the filter elements should be linked by OR operator and the value of
                each element is able to accept multiple values delimited by commas
                """
                values = ' OR '.join([item for item in val.split(',')])
                filter_str = '{}:({}) '.format(key, values)
                search_criterias['query']['query_string']['query'] += ' AND ({})'.format(filter_str)

         # handle pagination if parameters provided
        if kwargs.get('page', None):
            search_criterias.update({'from': (kwargs.get('page')-1)*kwargs.get('page_size', 30)})
            search_criterias.update({'size': kwargs.get('page_size', 30)})  

        result = json.loads(requests.post(url=url,
                                          data=json.dumps(search_criterias),
                                          headers={'Content-Type':'application/json'}).__dict__.get('_content'))

        if result.get('status', 200) != 404:
            return result.get('hits')
        else:
            return False

    def update(self, document, index, type, id):
        url = '{}/{}/{}/{}'.format(self._endpoint, index, type, id)
        if requests.put(url=url,
                        data=json.dumps(document),
                        headers={'Content-Type':'application/json'}).__dict__.get('status_code') == 200:
            return True
        else:
            return False

    def put_documents(self, documents, id_name, index, type):
        url = '{}/{}/{}/_bulk?pretty'.format(self._endpoint, index, type)
        payload = ''
        for doc in documents:
            payload += json.dumps({ "index":  { "_index": index, "_type": type, "_id": doc.get(id_name) }})
            payload += '\n'
            payload += json.dumps(doc)
            payload += '\n'

        if requests.put(url=url,
                        data=payload,
                        headers={'Content-Type':'application/json'}).__dict__.get('status_code') == 200:
            return True
        else:
            return False

    def delete_document(self, index, type, id):
        url = '{}/{}/{}/{}'.format(self._endpoint, index, type, id)
        if requests.delete(url=url).__dict__.get('status_code') == 200:
            return True
        else:
            return False

    def get_index(self, index):
        url = '{}/{}/_search'.format(self._endpoint, index)
        response = requests.get(url=url).__dict__

        if response.get('status', 404) == 404:
            return False

        return json.loads(response.get('_content')).get('hits')

    def put_index(self, index):
        if requests.put(url=''.join([self._endpoint, '/', index]),
                        headers={'Content-Type':'application/json'}).__dict__.get('status_code') == 200:
            return True
        else:
            return False

    def delete_index(self, index):
        if requests.delete(url=''.join([self._endpoint, '/', index])).__dict__.get('status_code') == 200:
            return True
        else:
            return False

    def get_type(self, index, type):
        url = '{}/{}/{}/_search'.format(self._endpoint, index, type)

        response = requests.get(url=url).__dict__

        if response.get('status', 404) == 404:
            return False

        return json.loads(response.get('_content')).get('hits')

class CloudSearch(Search):
    """
    This is for the implementation of AWS CloudSearch which built on top of Solr
    """
    def __init__(self, domain, endpoint):
        super().__init__(domain, endpoint)

    def search(self, keywords, index='', type=''):
        pass

    def update(self, document):
        pass

    def put_documents(self, documents, index, type):
        pass

    def get_index(self):
        pass

    def put_index(self, index):
        pass

    def delete_index(self, index):
        pass

    def get_type(self):
        pass


import json

class Policy(object):
    def __init__(self, version):
        self._policy = {'Version': version,
                        'Statement': []}

    @property
    def policy(self):
        return json.dumps(self._policy)

    def add_item(self, effect, actions, resources):
        raise NotImplementedError()

class IoTPolicy(Policy):
    def __init__(self, version='2012-10-17'):
        super().__init__(version)
    
    def add_item(self, effect, actions, resources):
        self._policy['Statement'].append({'Effect':effect,
                                          'Action':actions,
                                          'Resource':resources})


"""
This module contains subscriber role to AWS services
"""


class Subscriber(object):
    """
    Subscribe messages from SQS
    """
    @staticmethod
    def subscribe_to_queue(transport, delete_messages=True):
        """
        Subscribe to a queue
        """
        return transport.receive_messages(delete_messages)

    @staticmethod
    def subscribe_to_cache(transport):
        """
        Subscribe to AWS caching service
        """
        return transport.receive_messages()



import os
import six
from pynamodb.attributes import NumberAttribute, UTCDateTimeAttribute, UnicodeAttribute
from pynamodb.models import Model
from pynamodb.models import MetaModel
from pynamodb.connection.util import pythonic

def table_name(name):
    return name + '_' + os.environ.get('AWS_ENV', 'DEV').lower()

def table_region():
    return os.environ.get('AWS_DEFAULT_REGION', 'ap-southeast-1')

class MetaBase(MetaModel):
    def __new__(cls, name, bases, d, **kwargs):
        d['Meta'] = type('Meta', (), {
            'table_name': table_name(pythonic(name)),
            'region': table_region(),
            'read_capacity_units': 5,
            'write_capacity_units': 5
        })
        return MetaModel.__new__(cls, name, bases, d, **kwargs)

class ModelBase(Model):
    asset_manager_id = NumberAttribute(hash_key=True)
    created_by = UnicodeAttribute(null=False)
    created_time = UTCDateTimeAttribute(null=False)
    updated_by = UnicodeAttribute(null=False)
    updated_time = UTCDateTimeAttribute(null=False)

AMaaSModel = six.with_metaclass(MetaBase, ModelBase)





import json
import os
import pymysql
from boto3 import resource
from boto3.dynamodb.conditions import Key

class DynamoDBModelMixin(object):  
    def __str__(self):
        return json.dumps(self.__dict__)

    def to_json(self):
        return json.dumps(self.__dict__)

    def to_dict(self):
        return self.__dict__

class AssetModel(DynamoDBModelMixin):
    def __init__(self, asset_id, json_str):
        self.asset_id = asset_id
        self.json_str = json_str

class TransitionalAssetListModel(DynamoDBModelMixin):
    def __init__(self, batch_key, assets, retry_count=0):
        self.batch_key = batch_key
        self.assets = assets
        self.retry_count = retry_count

class EODExecutionModel(DynamoDBModelMixin):
    def __init__(self, batch_key, process_status):
        self.batch_key = batch_key
        self.process_status = process_status

class BatchExecutionModel(DynamoDBModelMixin):
    def __init__(self, batch_key, process_status, retry_count=0):
        self.batch_key = batch_key
        self.process_status = process_status
        self.retry_count = retry_count

class DynamoDB(object):
    def __init__(self, table_name, allowed_type=None):
        self._dynamodb_resource = resource('dynamodb')
        self._table = self._dynamodb_resource.Table(table_name)
        self._allowed_type = allowed_type

    def add(self, item):
        if isinstance(item, self._allowed_type):
            return self._table.put_item(Item=item.to_dict())
        else:
            raise TypeError('Invalid argument. {} object is expected'.format(self._allowed_type))

    def add_batch(self, items):
        if isinstance(items, list):
            new_list = list(filter(lambda item: isinstance(item, self._allowed_type), items))
            if len(items) != len(new_list):
                raise TypeError('Invalid argument. {} object is expected for all list elements'.format(self._allowed_type))
            else:
                new_list = []
                for item in items:
                    new_list.append({'PutRequest': {'Item':item.to_dict()}})

                if self._dynamodb_resource.batch_write_item(RequestItems={self._table.name: new_list}).get('UnprocessedItems'):
                    raise RuntimeError('Unable to process the batch insertion process.')
                else:
                    return True
        else:
            raise TypeError('Invalid argument. List object is expected')

    def get(self, key, value):
        return self._table.get_item(Key={key: value}).get('Item', None)

    def get_all(self, filter_key=None, filter_value=None, partial_match=False):
        """
        Perform a scan operation on table.
        Can specify filter_key (col name) and its value to be filtered.
        """

        if filter_key and filter_value:
            filtering_exp = Key(filter_key).eq(filter_value) if not partial_match else Key(filter_key).begins_with(filter_value)
            response = self._table.scan(FilterExpression=filtering_exp)
        else:
            response = self._table.scan()

        return response.get('Items', [])

    def update(self, key, value, update_key, update_value):
        # to prevent it from creating new item if key-value pair not found.
        # it should be handled by add function
        if self.get(key, value):
            return self._table.update_item(Key={key: value},
                                           UpdateExpression="SET {} = :updated".format(update_key),                   
                                           ExpressionAttributeValues={':updated': update_value},
                                           ReturnValues='UPDATED_NEW').get('Attributes', None)
        else:
            raise KeyError('No key-value pair found.')

    def delete(self, key, value):
        self._table.delete_item(Key={key: value})
        if self.get(key, value):
            raise KeyError('No key-value pair found and deleted')
        else:
            return True

    def purge(self):
        try:
            items = self._table.scan().get('Items', [])

            for item in items:
                self._table.delete_item(Key={list(item.keys())[0]: list(item.values())[0]})
        except Exception as e:
            raise e



import os
import logging
from typing import List
from amaasinfra.authorization.token_utils import TokenAttribute, unpack_token
from amaasinfra.authorization.user_permission import UserPermission
from amaasinfra.infra.dynamodb.asset_manager_relationship import AssetManagerRelationship
from amaasinfra.infra.dynamodb.book_permission import BookPermission

class AMaaSUser(object):
    def __init__(self, token: str) -> None:
        self.logger = logging.getLogger()
        attributes = unpack_token(token)
        user_amid = attributes.get(TokenAttribute.asset_manager_id.value)
        self.asset_manager_id = int(user_amid) if user_amid else None
        self.username = attributes.get(TokenAttribute.username.value)
        self.email = attributes.get(TokenAttribute.email.value)
        self.firstname = attributes.get(TokenAttribute.first_name.value, '')
        self.lastname = attributes.get(TokenAttribute.last_name.value, '')
        # amid 0-9 are reserved for system users
        self.system_user = self.asset_manager_id < 10 if self.asset_manager_id else False
        self.relationships: List[AssetManagerRelationship] = []
        self.data_sources: List[int] = []
        self.asset_manager_permissions: List[int] = []
        self.book_permissions: List[BookPermission] = []
        self.environment = os.environ.get('AWS_ENV')

    def load_relationships(self) -> 'AMaaSUser':
        if self.asset_manager_id is None:
            return self

        self.relationships = [rel for rel in AssetManagerRelationship
                            .related_id_index.query(self.asset_manager_id, relationship_status__eq='Active')]
        return self

    def load_data_sources(self) -> 'AMaaSUser':
        if self.asset_manager_id is None:
            return self

        if not self.relationships:
            self.load_relationships()

        asset_manager_ids = [rel.asset_manager_id for rel in self.relationships]
        data_providers = set()
        for asset_manager_id in asset_manager_ids:
            data_providers |= {rel.related_id
                            for rel in AssetManagerRelationship.query(hash_key=asset_manager_id,
                                                                      relationship_type__eq='Data Provider')}
        self.data_sources = list(data_providers)
        return self

    def load_asset_manager_permissions(self) -> 'AMaaSUser':
        if self.asset_manager_id is None:
            return self

        if not self.relationships:
            self.load_relationships()

        if not self.data_sources:
            self.load_data_sources()

        permissions = set()
        permissions |= {self.asset_manager_id}
        permissions |= {rel.asset_manager_id for rel in self.relationships}
        permissions |= {data_source for data_source in self.data_sources}
        self.asset_manager_permissions = list(permissions)

        return self

    def load_book_permissions(self) -> 'AMaaSUser':
        if self.asset_manager_id is None:
            return self

        self.book_permissions = [perm for perm in
                                 BookPermission.user_id_index.query(self.asset_manager_id,
                                                                    permission_status__eq='Active')]
        return self

    def check_asset_manager_permissions(self, asset_manager_id_claims: List[int]) -> (bool, List[int]):
        """
        Returns the asset_manager_ids that the user does not have access to
        """
        error_asset_manager_ids = [int(asset_manager_id) for asset_manager_id in asset_manager_id_claims
                                   if int(asset_manager_id) not in self.asset_manager_permissions]
        return error_asset_manager_ids


    def check_book_permissions(self, asset_manager_id, book_id, permission) -> None:
        if self.book_permissions is None:
            self.load_book_permissions()
        for book_permission in self.book_permissions:
            if (book_permission.book_id == book_id or book_permission.book_id == '*') and book_permission.asset_manager_id == asset_manager_id:
                if permission == 'write' and book_permission.permission == 'read':
                    raise TypeError('User %s does not have requested %s access to the book: %s'%(self.asset_manager_id, permission, book_id))
                return None
        raise TypeError('User %s does not have requested %s access to the book: %s'%(self.asset_manager_id, permission, book_id))


from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.attributes import UnicodeAttribute, BooleanAttribute
from amaasinfra.infra.dynamodb.amaas_model import AMaaSModel

class WorkflowContext(AMaaSModel):
    workflow_id = UnicodeAttribute(range_key=True)
    workflow_type = UnicodeAttribute(null=False)
    token = UnicodeAttribute(null=False)
    context = UnicodeAttribute()


from amaasinfra.infra.dynamodb.amaas_model import AMaaSModel
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.attributes import UnicodeAttribute, NumberAttribute

class RelatedIdIndex(GlobalSecondaryIndex):
    """
    Index for RelatedId.  
    Allows us to query for all relationships of a given RelatedId without 
    specifying the asset_manager_id partition key.
    """
    class Meta:
        index_name = "related_id_index"
        read_capacity_units = 5
        write_capacity_units = 5
        projection = AllProjection()

    related_id = NumberAttribute(hash_key=True)

class AssetManagerRelationship(AMaaSModel):
    related_id_index = RelatedIdIndex()
    related_id = NumberAttribute(range_key=True)
    relationship_type = UnicodeAttribute(null=False)
    relationship_status = UnicodeAttribute(null=False)
    
    


from amaasinfra.infra.dynamodb.amaas_model import AMaaSModel
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.attributes import UnicodeAttribute, NumberAttribute

class UserIdIndex(GlobalSecondaryIndex):
    class Meta:
        index_name = 'user_id_index'
        read_capacity_units = 5
        write_capacity_units = 5
        projection = AllProjection()
    user_asset_manager_id =  NumberAttribute(hash_key=True)

class BookPermission(AMaaSModel):
    book_id = UnicodeAttribute(range_key=True)
    user_id_index = UserIdIndex()
    user_asset_manager_id = NumberAttribute(null=False)
    permission_status = UnicodeAttribute(null=False)
    permission = UnicodeAttribute(null=False)



from datetime import datetime

from amaasinfra.communication.queue_publisher import QueuePublisher


class Publisher(object):
    """
    Publisher class is used to publish monitor messages, cache and
    other logs to the SQS
    """
    @staticmethod
    def publish_to_queue(message, transport, message_attributes=None):
        """
        This static method is used to passing messages to the SQS service
        """
        if not message:
            return False
        if message_attributes is None:
            message_attributes = {'date': {'StringValue': str(datetime.now()),
                                           'DataType': 'String'}}
        return QueuePublisher.publish_item(transport,
                                           message,
                                           message_attributes)

    @staticmethod
    def publish_to_cache():
        pass


from amaasinfra.infra.dynamodb.amaas_model import AMaaSModel
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.attributes import UnicodeAttribute, BooleanAttribute

class AssetClassIndex(GlobalSecondaryIndex):
    class Meta:
        index_name = 'asset_class_index'
        read_capacity_units = 5
        write_capacity_units = 5
        projection = AllProjection()

    asset_class = UnicodeAttribute(hash_key=True)

class Asset(AMaaSModel):
    asset_id = UnicodeAttribute(range_key=True)
    asset_class_index = AssetClassIndex()
    asset_class = UnicodeAttribute(null=False)
    asset_type = UnicodeAttribute(null=False)
    asset_status = UnicodeAttribute(null=False)
    fungible = BooleanAttribute()
    hashcode = UnicodeAttribute(null=False)

    


import click
import os
import pymysql
from configparser import ConfigParser, NoSectionError
from amaasinfra.config.amaas_config import AMaaSConfig

class DatabaseSetter(object):
    def __init__(self):
        pass

    def recreate(self, schema, environment, db_config={}):
        environment = environment.lower()
        config = AMaaSConfig()
        if environment in ['dev', 'staging']:
            click.echo(f"Using {environment} Settings")
            server = config.get_config_value(environment, 'db_server')
            username = config.get_config_value(environment, 'db_username')
            password = config.get_config_value(environment, 'db_password')
        elif environment == 'automation':
            click.echo(f"Using automation Settings")
            server = db_config.get('db_server')
            username = db_config.get('db_username')
            password = db_config.get('db_password')
        else:
            click.echo("Using Local Settings")
            server = "localhost"
            username = "root"
            password = "amaas"
        #connect to the server and create database
        try:
            click.echo("Connecting to the server")
            conn = pymysql.connect(server, user=username, passwd=password, charset='utf8')
            conn.cursor().execute("SET time_zone = '+00:00';")
            click.echo("Creating schema: " + schema)
            conn.cursor().execute("DROP DATABASE IF EXISTS "+ schema +";")
            conn.cursor().execute("CREATE DATABASE "+schema+" CHARACTER SET utf8 COLLATE utf8_unicode_ci;")
            conn.cursor().execute("USE "+schema+";")
            click.echo("Schema " + schema + " created successfully.")
        except (pymysql.InternalError, pymysql.OperationalError):
            raise Exception('Error Creating DB and Connection')

        automated = True if environment == 'automation' else False

        if not automated:
            tables_path = os.path.join(os.getcwd() + "/tables")
            data_path = os.path.join(os.getcwd() + "/data")
        else:
            tables_path = db_config.get('table_path')
            data_path = db_config.get('data_path')

        files = [(f, tables_path) for f in os.listdir(tables_path)]
        if os.path.exists(data_path):
            files += [(f, data_path) for f in os.listdir(data_path)]
        # Execute sql files
        try:
            for file, path in files:
                if automated and file in db_config.get('ignore'):
                    continue
                file_path = os.path.join(path, file)
                with open(file_path) as sql_file:
                    click.echo("Reading: " + file_path)
                    content = sql_file.readlines()
                exe_line = ''
                for line in content:
                    exe_line += line.partition('#')[0].rstrip()
                    if exe_line != '' and str(exe_line[-1]) == ';':
                        click.echo("Executing: " + exe_line)
                        conn.cursor().execute(exe_line)
                        exe_line = ''
                click.echo("Ran SQL: " + file[:-4] + " successfully")
            conn.commit()
        except pymysql.OperationalError:
            raise Exception('Error executing mysql files in tables')



"""
This lambda is used in Step Functions to invoke the lambdas in Singapore region that
performs the actual logic.
This is a temporary solution until Step Functions is available in Singapore region.
Currently this expects the lambdas to be invoked to specify the next step in the state machine.
Not ideal since it makes the specific steps aware of the execution chain but it greatly simplifies
the deployment
"""
import boto3
import logging
import json

logger = logging.getLogger()

def handler(event, context):
    next_lambda = event.get('next')
    # make the invoked lambda explicitly set this
    event['next'] = None

    if not next_lambda:
        error = 'The lambda function to invoke was not specified.'
        logger.error(error)
        raise RuntimeError(error)

    logger.info('Invoking {}'.format(next_lambda))
    client = boto3.client('lambda', region_name='ap-southeast-1')
    response = client.invoke(FunctionName=next_lambda, Payload=json.dumps(event))

    result = response.get('Payload')
    payload = result.read()
    # error = response.get('FunctionError')
    # error_message = error.read()

    # logger.info(error_message)
    # if error_message:
    #     raise RuntimeError(error_message)
    # todo: handle errors here
    return json.loads(payload)



import boto3

class IoTData(object):
    def __init__(self):
        self.client = boto3.client('iot-data')

    def publish(self, topic, payload, qos=1):
        """The message will be sent at least once"""
        return self.client.publish(
            topic=topic,
            qos=qos,
            payload=payload
        )